# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=invalid-name
# pylint: disable=global-statement
# pylint: disable=bare-except,broad-except
import argparse
import glob
import json
import time
from masnet import get_version, set_verbose, set_debug, set_working_dir
from masnet import debug, get_path
from masnet import save_graph, percent_progress
from masnet import save_labels, load_labels

# pylint: disable=too-many-statements
# pylint: disable=too-many-locals
def main():
    print('masnet v%s' % get_version())
    parser = argparse.ArgumentParser(prog='masnet.generate',
                                     description='',
                                     epilog='')

    parser.add_argument('-d', '--dir',
                        help='use specified directory for files',
                        required=False)

    parser.add_argument('--debug',
                        help='enables debug logging',
                        action='store_true',
                        required=False,
                        default=False)

    parser.add_argument('-v', '--verbose',
                        help='enable verbose logging, mostly for development',
                        action='store_true',
                        required=False,
                        default=False)

    args = parser.parse_args()
    set_debug(args.debug)
    set_verbose(args.verbose)
    debug(str(args))
    set_working_dir(args.dir)

    # node_id -> label (domain)
    id2label = {}
    # label -> node_id
    label2id = {}
    # node_id -> adjlist of this node
    adjlist = {}
    # node_id -> list of peers from json (domain names)
    peers = {}
    num_nodes = 0
    num_links = 0
    start = time.time()

    def print_status():
        elapsed = time.time() - start
        hours = int(elapsed / 3600)
        minutes = int((elapsed - hours * 3600) / 60)
        seconds = elapsed - minutes * 60 - hours * 3600
        print('N:%08d L:%08d ' \
              'e:%02d:%02d:%02d ' % (num_nodes, num_links,
                                     hours, minutes, seconds))


    try:
        # load domains from peers.json files
        # this is the best way because otherwise exclusion and errors has to
        # be re-checked, computationally expensive
        # we already know these passed exclusion and has peers info
        print('creating the graph from peers...')
        last_status = time.time() - 10
        for file_name in glob.glob(get_path('*.peers.json')):
            if (time.time() - last_status) > 1:
                print_status()
                last_status = time.time()
            with open(file_name, 'r') as f:
                doc = json.load(f)
                # reading the domain from the file, so this is the actual
                # domain
                domain = doc['domain']
                node_id = num_nodes
                num_nodes = num_nodes + 1
                label2id[domain] = node_id
                id2label[node_id] = domain
                peers[node_id] = doc['peers']
                # adjlist is a set, hence no multiple links
                adjlist[node_id] = set()
        for node_id, node_peers in peers.items():
            if (time.time() - last_status) > 1:
                print_status()
                last_status = time.time()
            for peer_name in node_peers:
                # there should be no need for None and len=0 checks
                # but I saw such data can be returned from peers api call
                # so clean it up
                if peer_name is None:
                    continue
                if len(peer_name.strip()) == 0:
                    continue
                # filter if peer is not a known node
                # Mastodon peers usually contain many strange domains:
                # - private IP addresses
                # - malicious domains
                # - not working domains
                if peer_name in label2id:
                    peer_node_id = label2id[peer_name]
                    # do not allow self-loops
                    if peer_node_id != node_id:
                        adjlist[node_id].add(peer_node_id)
                        num_links = num_links + 1

        del peers

        print_status()
        print('graph created.')

        save_labels(id2label,
                    get_path('mastodon.labels'))
        print('labels saved.')

        del id2label
        del label2id

        print('saving graphs...')
        save_graph(adjlist,
                   False,
                   get_path('mastodon.networkit.undirected'),
                   progressfn=percent_progress)
        save_graph(adjlist,
                   True,
                   get_path('mastodon.networkit.directed'),
                   progressfn=percent_progress)
        print('graphs saved.')

    except KeyboardInterrupt:
        pass

    print('bye.')

if __name__ == '__main__':
    main()



# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=invalid-name
# pylint: disable=global-statement
# pylint: disable=bare-except,broad-except
import argparse
import asyncio
import json
import signal
import time
import traceback
import aiodns
import aiofile
import aiohttp
from tabulate import tabulate
from masnet import get_peers_file_path, get_error_file_path, load_exclusion
from masnet import get_version, set_verbose, set_working_dir, debug, verbose
from masnet import set_debug, is_excluded, get_path, is_debug, is_verbose
from masnet import get_excluded_patterns


RUN = True
num_nodes = 0
num_links = 0
num_errors = 0
num_skips = 0
num_timeouts = 0


# pylint: disable=unused-argument
def signal_handler(sig, frame):
    global RUN
    RUN = False


async def save_error(domain, error):
    async with aiofile.async_open(get_path('masnet.download.errors'),
                                  'a') as afp:
        await afp.write('%s %s\n' % (domain, error))


async def save_skip(domain):
    async with aiofile.async_open(get_path('masnet.download.skips'),
                                  'a') as afp:
        await afp.write('%s\n' % domain)


async def save_time(domain, download_time):
    async with aiofile.async_open(get_path('masnet.download.times'),
                                  'a') as afp:
        await afp.write('%s %d\n' % (domain, download_time))


async def save_visit(domain):
    async with aiofile.async_open(get_path('masnet.download.visits'),
                                  'a') as afp:
        await afp.write('%s\n' % domain)


# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
async def fetch(session, domain, timeout, scheduled, q):
    global num_errors, num_nodes, num_skips, num_links, num_timeouts
    url = 'https://%s/api/v1/instance/peers' % domain
    try:
        start = time.time()
        async with session.get(url, timeout=timeout) as resp:
            if resp.status == 200:
                resp_as_text = await resp.text()
                await save_time(domain,
                                int((time.time() - start) * 1000))
                peers = json.loads(resp_as_text)
                if peers is None:
                    num_errors = num_errors + 1
                    await save_error(domain, 'peers is None')
                elif not isinstance(peers, list):
                    num_errors = num_errors + 1
                    await save_error(domain, 'peers is not list')
                else:
                    doc = {'domain': domain,
                           'peers': peers}
                    async with aiofile.async_open(get_path('%s.peers.json' % domain),
                                                  'w') as afp:
                        await afp.write(json.dumps(doc))
                    num_nodes = num_nodes + 1
                    await save_visit(domain)
                    for peer in peers:
                        if peer is None:
                            continue
                        if len(peer.strip()) == 0:
                            continue
                        if is_excluded(peer):
                            await save_skip(peer)
                            num_skips = num_skips + 1
                            continue
                        num_links = num_links + 1
                        if peer not in scheduled:
                            scheduled.add(peer)
                            await q.put(peer)
            else:
                num_errors = num_errors + 1
                await save_error(domain, 'HTTP: %d' % resp.status)
    except aiohttp.ClientError as e:
        num_errors = num_errors + 1
        error = e.__class__.__name__
        await save_error(domain, error)
    except aiodns.error.DNSError as e:
        num_errors = num_errors + 1
        error = e.__class__.__name__
        await save_error(domain, error)
    except asyncio.exceptions.TimeoutError as e:
        num_errors = num_errors + 1
        num_timeouts = num_timeouts + 1
        error = e.__class__.__name__
        await save_error(domain, error)
    except asyncio.CancelledError as e:
        num_errors = num_errors + 1
        error = e.__class__.__name__
        await save_error(domain, error)
    except json.decoder.JSONDecodeError as e:
        num_errors = num_errors + 1
        error = e.__class__.__name__
        await save_error(domain, error)
    except UnicodeError as e:
        num_errors = num_errors + 1
        error = e.__class__.__name__
        await save_error(domain, error)
    except Exception as e:
        num_errors = num_errors + 1
        error = e.__class__.__name__
        await save_error(domain, error)
        if is_verbose():
            traceback.print_exc()


# pylint: disable=too-many-arguments
# pylint: disable=too-many-branches
# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
async def download(start_domain,
                   timeout,
                   num_tasks):

    start = time.time()
    barrier = asyncio.Semaphore(num_tasks)
    scheduled = set()
    q = asyncio.Queue()
    resolver = aiohttp.resolver.AsyncResolver(nameservers=['8.8.8.8',
                                                           '8.8.4.4'])
    conn = aiohttp.TCPConnector(limit=2*num_tasks,
                                force_close=True, # no need for keep-alive
                                use_dns_cache=False,
                                resolver=resolver)
    async with aiohttp.ClientSession(connector=conn,
                                     timeout=timeout) as session:
        domain_counts = {}
        scheduled.add(start_domain)
        q.put_nowait(start_domain)
        last = time.time()
        signal.signal(signal.SIGINT, signal_handler)
        cnt = 0
        cnt_threshold = 5
        top5 = []
        def print_status():
            elapsed = time.time() - start
            hours = int(elapsed / 3600)
            minutes = int((elapsed - hours * 3600) / 60)
            seconds = elapsed - minutes * 60 - hours * 3600
            print('q:%06d a:%06d s:%06d ' \
                  'N:%06d L:%09d ' \
                  'e:%08d s:%08d to:%04d ' \
                  't:%02d:%02d:%02d' % (q.qsize(), len(asyncio.all_tasks()),
                                        len(scheduled),
                                        num_nodes, num_links,
                                        num_errors, num_skips, num_timeouts,
                                        hours, minutes, seconds), flush=True)
            verbose(tabulate(top5))
        global RUN
        while RUN and (cnt < cnt_threshold):
            if (time.time() - last) > 1:
                print_status()
                last = time.time()
                if len(asyncio.all_tasks()) == 1:
                    cnt = cnt + 1
                else:
                    cnt = 0

            if q.qsize() > 0:
                if not barrier.locked():
                    async with barrier:
                        domain = await q.get()
                        debug(domain)
                        if is_verbose():
                            parts = domain.split('.')
                            if len(parts) > 2:
                                count_key = '.'.join(parts[-(len(parts)-1):])
                                domain_counts[count_key] = domain_counts.get(count_key,
                                                                             0) + 1
                                top5 = list(sorted(domain_counts.items(),
                                                   key=lambda t: t[1], reverse=True))[:5]
                        asyncio.create_task(fetch(session,
                                                  domain,
                                                  timeout,
                                                  scheduled,
                                                  q))
            # this is needed to give time to async loop operation
            await asyncio.sleep(0)

        print_status()

        if RUN:
            print('traversal finished. data is complete.')
        else:
            print('traversal terminated early. data is incomplete !!!')

        current_task = asyncio.current_task()

        print('cancelling tasks...')
        running_tasks = filter(lambda x: x != current_task,
                               asyncio.all_tasks())
        for task in running_tasks:
            try:
                task.cancel()
            except:
                pass

        print('waiting for tasks...')
        running_tasks = set(filter(lambda x: x != current_task,
                                   asyncio.all_tasks()))
        for task in running_tasks:
            try:
                await task
            except:
                pass
        print('bye.')


# pylint: disable=too-many-statements
def main():
    print('masnet v%s' % get_version())
    parser = argparse.ArgumentParser(prog='masnet.download',
                                     description='',
                                     epilog='')

    parser.add_argument('-d', '--dir',
                        help='use specified directory for files',
                        default=None,
                        required=False)

    parser.add_argument('--debug',
                        help='enables debug logging',
                        action='store_true',
                        required=False,
                        default=False)

    parser.add_argument('-v', '--verbose',
                        help='enable verbose logging, mostly for development',
                        action='store_true',
                        required=False,
                        default=False)

    parser.add_argument('-e', '--exclude-file',
                        help='exclude the domains and all of their ' \
                             'subdomains in the file specified',
                        required=False)

    parser.add_argument('-n', '--num-tasks',
                        help='use specified number of Python asyncio tasks for download',
                        type=int,
                        required=False,
                        default=100)

    parser.add_argument('-s', '--start-domain',
                        help='domains to start traversal from (default: mastodon.social)',
                        required=False,
                        default='mastodon.social')

    parser.add_argument('-t', '--timeout',
                        help='use specified number of seconds for timeout ' \
                             '(default: system default)',
                        type=int,
                        required=False,
                        default=60)

    args = parser.parse_args()
    set_debug(args.debug)
    set_verbose(args.verbose)
    debug(str(args))
    load_exclusion(args.exclude_file)
    set_working_dir(args.dir)

    # truncate fixed name output files
    with open(get_path('masnet.download.errors'), 'w'):
        pass
    with open(get_path('masnet.download.visits'), 'w'):
        pass
    with open(get_path('masnet.download.skips'), 'w'):
        pass
    with open(get_path('masnet.download.times'), 'w'):
        pass

    asyncio.run(download(args.start_domain,
                         args.timeout,
                         args.num_tasks))


if __name__ == '__main__':
    main()



# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=invalid-name
# pylint: disable=global-statement
# pylint: disable=bare-except,broad-except
import argparse
import time
import matplotlib.pyplot as plt
import networkit as nk
import networkit.vizbridges as nkvz
import numpy
from tabulate import tabulate, SEPARATING_LINE
from masnet import get_version, set_verbose, set_debug, debug, set_working_dir
from masnet import load_labels, get_path, pltpause

# pylint: disable=too-many-statements
# pylint: disable=too-many-locals
def main():
    print('masnet v%s' % get_version())
    parser = argparse.ArgumentParser(prog='masnet.generate',
                                     description='',
                                     epilog='')

    parser.add_argument('-d', '--dir',
                        help='use specified directory for files ' \
                             '(default: current directory)',
                        required=False)

    parser.add_argument('--debug',
                        help='enables debug logging',
                        action='store_true',
                        required=False,
                        default=False)

    parser.add_argument('-v', '--verbose',
                        help='enable verbose logging, mostly for development',
                        action='store_true',
                        required=False,
                        default=False)

    parser.add_argument('--directed',
                        help='use directed graph (default: use undirected)',
                        action='store_true',
                        required=False,
                        default=False)

    parser.add_argument('--card',
                        help='shows the network card',
                        action='store_true',
                        required=False,
                        default=False)

    parser.add_argument('--strongly-connected-components',
                        help='(experimental)',
                        action='store_true',
                        default=None,
                        required=False)

    parser.add_argument('--local-clustering-coefficients',
                        help='(experimental)',
                        action='store_true',
                        default=None,
                        required=False)

    parser.add_argument('--cut-clusters',
                        help='(experimental)',
                        action='store_true',
                        default=None,
                        required=False)

    parser.add_argument('--core-decomposition',
                        help='(experimental)',
                        action='store_true',
                        default=None,
                        required=False)

    parser.add_argument('--detect-communities',
                        help='(experimental)',
                        action='store_true',
                        default=None,
                        required=False)

    parser.add_argument('--prune',
                        help='(experimental)',
                        choices=['out-degree'],
                        default=None,
                        required=False)

    parser.add_argument('--prune-cutoff',
                        help='(experimental)',
                        type=int,
                        default=0,
                        required=False)

    parser.add_argument('--prune-out',
                        help='(experimental)',
                        default='mastodon.partition.graph',
                        required=False)

    parser.add_argument('--degree-centrality',
                        help='(experimental)',
                        action='store_true',
                        required=False,
                        default=False)

    args = parser.parse_args()
    set_debug(args.debug)
    set_verbose(args.verbose)
    debug(str(args))
    set_working_dir(args.dir)

    id2label = load_labels(get_path('mastodon.labels'))
    print('node labels loaded.')
    start = time.time()
    g = None
    if args.directed:
        print('reading directed graph into networkit...')
        g = nk.readGraph(get_path('mastodon.networkit.directed'),
                         nk.graphio.Format.NetworkitBinary)
        if not g.isDirected():
            print('mastodon.networkit.directed is undirected ?!')
            sys.exit(-1)
    else:
        print('reading undirected graph into networkit...')
        g = nk.readGraph(get_path('mastodon.networkit.undirected'),
                        nk.graphio.Format.NetworkitBinary)
        if g.isDirected():
            print('mastodon.networkit.undirected is directed ?!')
            sys.exit(-1)
    print('graph read in %.1f seconds.' % (time.time() - start))

    if args.card:

        print('generating the network card...')

        N = g.numberOfNodes()
        L = g.numberOfEdges()

        k_avg = 0
        k_min = N-1
        k_max = 0
        for node_id in g.iterNodes():
            k = g.degree(node_id)
            if k < k_min:
                k_min = k
            if k > k_max:
                k_max = k
            k_avg = k_avg + k
        k_avg = k_avg / N

        components_run = nk.components.ConnectedComponents(g)
        components_run.run()
        component_sizes = components_run.getComponentSizes()

        card = []
        card.append(['Name', 'Mastodon peers network'])
        card.append(['Kind', '%s, %s' % ('directed' if g.isDirected() else 'undirected',
                                         'weighted' if g.isWeighted() else 'unweighted')])
        card.append(['Nodes are', 'Mastodon instances'])
        card.append(['Links are', 'Peer relationship'])
        card.append(SEPARATING_LINE)
        card.append(['Number of nodes', N])
        card.append(['Number of links', L])
        card.append(['Degree*', '%.3f [%d, %d]' % (k_avg,
                                                   k_min,
                                                   k_max)])

        if g.isDirected():
            card.append(['Clustering', 'n/a'])
        else:
            lcc_run = nk.centrality.LocalClusteringCoefficient(g,
                                                            turbo=True)
            lcc_run.run()
            avg_clustering = sum(lcc_run.scores()) / N
            card.append(['Clustering', '%.3f' % avg_clustering])

        if len(component_sizes) == 1:

            diameter_run = nk.distance.Diameter(g,
                                                algo=nk.distance.DiameterAlgo.Exact)
            diameter_run.run()
            diamater = diameter_run.getDiameter()

            card.append(['Connected', 'Yes'])
            card.append(['Diameter', '%.3f' % diameter])
        else:
            comp_sizes = component_sizes.values()
            Nmaxcomp = max(comp_sizes)
            card.append(['Connected',
                         '%d components [%.1f%% in largest]' % (len(component_sizes),
                                                                100*Nmaxcomp/N)])
            Nmincomp = min(comp_sizes)
            Navgcomp = sum(comp_sizes) / len(comp_sizes)
            card.append(['Component size*', '%.1f [%d, %d]' % (Navgcomp,
                                                              Nmincomp,
                                                              Nmaxcomp)])
            card.append(['Diameter', 'n/a'])

            largest_component = components_run.extractLargestConnectedComponent(g,
                                                                                False)
            diameter_run = nk.distance.Diameter(largest_component,
                                                algo=nk.distance.DiameterAlgo.Exact)
            diameter_run.run()
            diameter = diameter_run.getDiameter()[0]
            card.append(['Largest component\'s diameter', '%d' % diameter])

        card.append(SEPARATING_LINE)
        card.append(['Data generating process', 'masnet.download, masnet.generate'])

        print(tabulate(card))

        print('*: avg [min, max]')

    elif args.strongly_connected_components:

        scc = nk.components.StronglyConnectedComponents(g)
        scc.run()
        print('# of strongly connected components: %d' % scc.numberOfComponents())
        print('component sizes: %s' % scc.getComponentSizes().values())

    elif args.local_clustering_coefficients:

        lcc = nk.centrality.LocalClusteringCoefficient(ug,
                                                       turbo=True)
        lcc.run()
        ranks = lcc.ranking()[:100]
        for rank in ranks:
            domain = id2label[rank[0]]
            local_clustering_coefficient = rank[1]
            print('%s %0.3f' % (domain, local_clustering_coefficient))

    elif args.cut_clusters:

        for alpha in [0, 0.25, 0.5, 0.75, 1]:
            cc = nk.community.CutClustering(g, alpha)
            cc.run()

    elif args.detect_communities:

        #cs = nk.community.detectCommunities(ug)
        cs = nk.community.detectCommunities(ug,
                                            algo=nk.community.PLM(ug, True))
        print(cs.subsetSizes())

    elif args.degree_centrality:

        dc = sorted(nk.centrality.DegreeCentrality(g,
                                                   normalized=False,
                                                   outDeg=True).run().scores(),
                    reverse=True)

        degrees, number_of_nodes = numpy.unique(dc, return_counts=True)

        plt.xscale('log')
        plt.xlabel('degree')
        plt.yscale('log')
        plt.ylabel('number of nodes')
        plt.title('degree distribution')
        plt.plot(degrees, number_of_nodes)
        plt.show(block=False)
        pltpause()

    elif args.core_decomposition:

        cd = nk.centrality.CoreDecomposition(ug)
        cd.run()
        scores = cd.scores()
        max_score = max(scores)
        print('max score: %s' % max_score)
        snodes = list()
        for node_id in g.iterNodes():
            if scores[node_id] == max_score:
                snodes.append(node_id)

        sg = nk.graphtools.subgraphFromNodes(ug, snodes)
        print(sg.numberOfNodes())
        print(sg.numberOfEdges())
        nk.writeGraph(sg,
                      'core-decomposition.gml',
                      nk.graphio.Format.GML)

    elif args.prune is not None:
        degrees = list()
        for node_id in g.iterNodes():
            kin = g.degreeIn(node_id)
            kout = g.degreeOut(node_id)
            degrees.append((kin, kout, node_id))
        degrees = sorted(degrees, key=lambda x: x[1], reverse=True)
        kept_nodes = list(map(lambda x: x[2], degrees[0:100]))
        pruned_graph = nk.graphtools.subgraphFromNodes(g, kept_nodes)
        nk.writeGraph(pruned_graph,
                      'mastodon.partition.gml',
                      nk.graphio.Format.GML)

    else:
        parser.print_help(sys.stdout)

if __name__ == '__main__':
    main()



# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
import gzip
import os
import re
from struct import pack, unpack
import time
try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources

__version__ = '0.3.1'

def get_version():
    return __version__

def get_path(file_name):
    if file_name is None:
        return DIR
    else:
        return os.path.join(DIR, file_name)

def get_peers_file_name(domain):
    return '%s.peers.json' % domain.replace('/', '_')

def get_peers_file_path(domain):
    return get_path(get_peers_file_name(domain))

def get_error_file_name(domain):
    return '%s.error' % domain.replace('/', '_')

def get_error_file_path(domain):
    return get_path(get_error_file_name(domain))

EXCLUDED_PATTERNS = list()

def load_exclusion(exclude_file):
    global EXCLUDED_PATTERNS
    def load_exclusion_file(f):
        for line in f:
            line = line.strip()
            if line.startswith('#'):
                continue
            if len(line) == 0:
                continue
            verbose(line)
            EXCLUDED_PATTERNS.append(re.compile(line))
    verbose('--- start of exclusion patterns ---')
    if exclude_file is None:
        with pkg_resources.path('masnet',
                                'default_exclusion_patterns') as ipath:
            verbose('loading internal exclusion list from: %s' % ipath)
            with open(ipath, 'r') as file:
                load_exclusion_file(file)
    else:
        verbose('loading external exclusion list from: %s' % exclude_file)
        with open(exclude_file, 'r') as file:
            load_exclusion_file(file)
    verbose('--- end of exclusion patterns ---')

def is_excluded(domain):
    for pattern in EXCLUDED_PATTERNS:
        if pattern.search(domain, re.IGNORECASE) is not None:
            debug('%s excluded due to: %s' % (domain, pattern))
            return True
    return False

def get_excluded_patterns():
    return EXCLUDED_PATTERNS

VERBOSE = False
def set_verbose(verbose):
    global VERBOSE
    VERBOSE = verbose

def is_verbose():
    return VERBOSE

def verbose(s):
    if VERBOSE:
        print(s, flush=True)

DEBUG = False
def set_debug(debug):
    global DEBUG
    DEBUG = debug

def is_debug():
    return DEBUG

def debug(s):
    if DEBUG:
        print(s, flush=True)

def pltpause():
    import matplotlib.pyplot as plt
    input('press enter to close plots...')
    plt.close('all')

DIR = None
def set_working_dir(wd):
    global DIR
    if wd is None:
        DIR = os.getcwd()
    else:
        DIR = os.path.abspath(wd)
        os.makedirs(DIR, exist_ok=True)


def percent_progress(x):
    if x == -1:
        percent_progress.last = 0
    elif x == 2:
        print('100%')
    elif (time.time() - percent_progress.last) > 1:
        print('%d%% ' % (x*100), end='', flush=True)
        percent_progress.last = time.time()


def save_graph(adjlists,
               create_directed,
               file_path,
               progressfn=None):
    if progressfn:
        progressfn(-1)
    import networkit as nk
    g = nk.Graph(n=len(adjlists),
                 directed=create_directed)
    # pylint: disable=consider-using-enumerate
    # not enumerating to preserve the order 0...n
    for i in range(0, len(adjlists)):
        if progressfn:
            progressfn(i/len(adjlists))
        adjlist = adjlists[i]
        for adj_vertex_id in sorted(adjlist):
            if create_directed:
                g.addEdge(i, adj_vertex_id)
            # prevent adding both u->v and v->u if undirected
            elif i > adj_vertex_id:
                g.addEdge(i, adj_vertex_id)

    # save graph
    nk.writeGraph(g,
                  file_path,
                  nk.Format.NetworkitBinary,
                  chunks=32,
                  NetworkitBinaryWeights=0) # 0=no weight
    if progressfn:
        progressfn(2)


def save_labels(id2labels, file_path):
    with open(get_path(file_path), 'w') as file:
        # pylint: disable=consider-using-enumerate
        # not enumerating to preserve the order 0...n
        for i in range(0, len(id2labels)):
            file.write('%s\n' % id2labels[i])


def load_labels(file_path):
    id2labels = {}
    i = 0
    with open(get_path(file_path), 'r') as file:
        for line in file:
            id2labels[i] = line.strip()
            i = i + 1
    return id2labels



# coding: utf-8
from setuptools import setup

# remove CircleCI link because it points to branch, not a tag etc.
def readme():
    lines = []
    with open('README.md') as f:
        while True:
            line = f.readline()
            if len(line) == 0:
                break
            # remove CircleCI
            if line.startswith('[![CircleCI](https'):
                continue
            lines.append(line)
    return ''.join(lines)

setup(long_description=readme())
